@@ -52,7 +52,7 @@ impl MetricsPingSubmitter for GleanMetricsPingSubmitter {
     fn submit_metrics_ping(&self, glean: &Glean, reason: Option<&str>, now: DateTime<FixedOffset>) {
         glean.submit_ping_by_name("metrics", reason);
         // Always update the collection date, irrespective of the ping being sent.
-        get_last_sent_time_metric().set(&glean, Some(now));
+        get_last_sent_time_metric().set(glean, Some(now));
     }
 }
 
@@ -84,7 +84,7 @@ pub fn schedule(glean: &Glean) {
     let submitter = GleanMetricsPingSubmitter {};
     let scheduler = GleanMetricsPingScheduler {};
 
-    schedule_internal(&glean, submitter, scheduler, now)
+    schedule_internal(glean, submitter, scheduler, now)
 }
 
 /// Tells the scheduler task to exit quickly and cleanly.
@@ -101,21 +101,21 @@ fn schedule_internal(
     now: DateTime<FixedOffset>,
 ) {
     let last_sent_build_metric = get_last_sent_build_metric();
-    if let Some(last_sent_build) = last_sent_build_metric.get_value(&glean, INTERNAL_STORAGE) {
+    if let Some(last_sent_build) = last_sent_build_metric.get_value(glean, INTERNAL_STORAGE) {
         // If `app_build` is longer than StringMetric's max length, we will always
         // treat it as a changed build when really it isn't.
         // This will be externally-observable as InvalidOverflow errors on both the core
         // `client_info.app_build` metric and the scheduler's internal metric.
         if last_sent_build != glean.app_build {
-            last_sent_build_metric.set(&glean, &glean.app_build);
+            last_sent_build_metric.set(glean, &glean.app_build);
             log::info!("App build changed. Sending 'metrics' ping");
-            submitter.submit_metrics_ping(&glean, Some("upgrade"), now);
+            submitter.submit_metrics_ping(glean, Some("upgrade"), now);
             scheduler.start_scheduler(submitter, now, When::Reschedule);
             return;
         }
     }
 
-    let last_sent_time = get_last_sent_time_metric().get_value(&glean, INTERNAL_STORAGE);
+    let last_sent_time = get_last_sent_time_metric().get_value(glean, INTERNAL_STORAGE);
     if let Some(last_sent) = last_sent_time {
         log::info!("The 'metrics' ping was last sent on {}", last_sent);
     }
@@ -137,7 +137,7 @@ fn schedule_internal(
     } else if now > now.date().and_hms(SCHEDULED_HOUR, 0, 0) {
         // Case #2
         log::info!("Sending the 'metrics' ping immediately, {}", now);
-        submitter.submit_metrics_ping(&glean, Some("overdue"), now);
+        submitter.submit_metrics_ping(glean, Some("overdue"), now);
         scheduler.start_scheduler(submitter, now, When::Reschedule);
     } else {
         // Case #3
@@ -198,31 +198,42 @@ fn start_scheduler(
             loop {
                 let dur = when.until(now);
                 log::info!("Scheduling for {:?} after {}, reason {:?}", dur, now, when);
-                let result = condvar.wait_timeout_while(cancelled_lock.lock().unwrap(), dur, |cancelled| !*cancelled);
-                now = local_now_with_offset().0;
-                match result {
-                    Err(err) => {
-                        log::warn!("Condvar wait failure, {}", err);
-                        break;
-                    }
-                    Ok((cancelled, wait_result)) => {
-                        if *cancelled {
-                            log::info!("Metrics Ping Scheduler cancelled. Exiting.");
+                let mut timed_out = false;
+                {
+                    match condvar.wait_timeout_while(cancelled_lock.lock().unwrap(), dur, |cancelled| !*cancelled) {
+                        Err(err) => {
+                            log::warn!("Condvar wait failure. MPS exiting. {}", err);
                             break;
-                        } else if wait_result.timed_out() {
-                            log::info!("Time to submit our metrics ping, {:?}", when);
-                            let glean = crate::global_glean().expect("Global Glean not present when trying to send scheduled 'metrics' ping?!").lock().unwrap();
-                            submitter.submit_metrics_ping(&glean, Some(when.reason()), now);
-                            when = When::Reschedule;
-                        } else {
-                            // This should be impossible. `cancelled_lock` is acquired, and
-                            // `!*cancelled` is checked by the condvar before it is allowed
-                            // to return from `wait_timeout_while` (I checked).
-                            // So `Ok(_)` implies `*cancelled || wait_result.timed_out`.
-                            log::warn!("Spurious wakeup of the MPS condvar should be impossible.");
+                        }
+                        Ok((cancelled, wait_result)) => {
+                            if *cancelled {
+                                log::info!("Metrics Ping Scheduler cancelled. Exiting.");
+                                break;
+                            } else if wait_result.timed_out() {
+                                // Can't get the global glean while holding cancelled's lock.
+                                timed_out = true;
+                            } else {
+                                // This should be impossible. `cancelled_lock` is acquired, and
+                                // `!*cancelled` is checked by the condvar before it is allowed
+                                // to return from `wait_timeout_while` (I checked).
+                                // So `Ok(_)` implies `*cancelled || wait_result.timed_out`.
+                                log::warn!("Spurious wakeup of the MPS condvar should be impossible.");
+                            }
                         }
                     }
                 }
+                // Safety:
+                // We are okay dropping the condvar's cancelled lock here because it only guards
+                // whether we're cancelled, and we've established that we weren't when we timed out.
+                // We might _now_ be cancelled at any time, in which case when we loop back over
+                // we'll immediately exit. But first we need to submit our "metrics" ping.
+                if timed_out {
+                    log::info!("Time to submit our metrics ping, {:?}", when);
+                    let glean = crate::global_glean().expect("Global Glean not present when trying to send scheduled 'metrics' ping?!").lock().unwrap();
+                    submitter.submit_metrics_ping(&glean, Some(when.reason()), now);
+                    when = When::Reschedule;
+                }
+                now = local_now_with_offset().0;
             }
         }).expect("Unable to spawn Metrics Ping Scheduler thread.")
 }
